"""The 'worktoy.meta' module provides the implementation of 'overload'."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Overload


def overload(*types) -> Overload:
  """Decorator function for overloading functions."""
  typeNames = [cls.__name__ for cls in types]
  info = """Overloading types: '%s'""" % ', '.join(typeNames)
  # print(info)
  return Overload(*types)



"""SpaceNum provides the namespace class used by the MetaNum metaclass."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Any, Self, TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False
  Self = object
  Any = object

from worktoy.keenum import Num
from worktoy.meta import AbstractNamespace, OverloadSpace, CallMeMaybe
from worktoy.parse import maybe
from worktoy.text import monoSpace

if TYPE_CHECKING:
  NumList = list[Num]


class SpaceNum(OverloadSpace):
  """SpaceNum provides the namespace class used by the MetaNum metaclass."""

  __num_entries__ = None
  __reserved_names__ = ['__num_entries__', '__init__', ]

  def _getNumEntries(self) -> NumList:
    """Get the Num entries."""
    return maybe(self.__num_entries__, [])

  def _addNumEntry(self, key: str, num: Num) -> None:
    """Add a Num entry."""
    existing = self._getNumEntries()
    Num.setPrivateValue(num, len(existing))
    Num.setPublicName(num, key)
    self.__num_entries__ = [*existing, num]

  def _validateName(self, name: str) -> str:
    """Validate the name."""
    if name in self.__reserved_names__:
      e = """Received a reserved name: '%s'!""" % name
      raise AttributeError(monoSpace(e))
    if name.lower() in self._getNumEntries():
      e = """Received a name that is already in use: '%s'!""" % name
      raise AttributeError(monoSpace(e))
    return name

  def __setitem__(self, key: str, value: object) -> None:
    """Set the key, value pair."""
    if not getattr(value, '__trust_me_bro__', False):
      key = self._validateName(key)
    if isinstance(value, Num):
      return self._addNumEntry(key, value)
    return OverloadSpace.__setitem__(self, key, value)

  def compile(self, ) -> dict:
    """Compile the namespace."""
    return {**OverloadSpace.compile(self),
            '__num_entries__': self._getNumEntries(),
            '__keenum_list__': [],
            '__keenum_dict__': {}}



"""TypeSig encapsulates type signatures and the functionality for
recognizing positional arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.parse import maybe

try:
  from typing import Any
except ImportError:
  Any = object

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

from worktoy.text import monoSpace, typeMsg

if TYPE_CHECKING:
  from typing import Union, Optional

  Number = Union[int, float, complex]
  MaybeInt = Optional[int]
  MaybeFloat = Optional[float]
  MaybeComplex = Optional[complex]
else:
  Number = object
  MaybeInt = object
  MaybeFloat = object
  MaybeComplex = object


class TypeSig:
  """TypeSig encapsulates type signatures and the functionality for
  recognizing positional arguments. """

  __type_signature__ = None
  __hash_value__ = None

  @classmethod
  def castInt(cls, num: Number) -> MaybeInt:
    """Casts the number to an integer."""
    if isinstance(num, int):
      return num
    if isinstance(num, float):
      if num.is_integer():
        return int(num)
      return None
    if isinstance(num, complex):
      if num.imag:
        return None
      return cls.castInt(num.real)
    return None

  @classmethod
  def castFloat(cls, num: Number) -> MaybeFloat:
    """Casts the number to a float."""
    if isinstance(num, int):
      return float(num)
    if isinstance(num, float):
      return num
    if isinstance(num, complex):
      if num.imag:
        return None
      return cls.castFloat(num.real)
    return None

  @classmethod
  def castComplex(cls, num: Number) -> MaybeComplex:
    """Casts the number to a complex number."""
    if isinstance(num, int):
      return cls.castFloat(num) + 0j
    if isinstance(num, float):
      return num + 0j
    if isinstance(num, complex):
      return num
    return None

  @classmethod
  def castArg(cls, arg: object, type_: type) -> object:
    """Casts the argument to the type. """
    if type_ is int:
      return cls.castInt(arg)
    if type_ is float:
      return cls.castFloat(arg)
    if type_ is complex:
      return cls.castComplex(arg)
    if isinstance(arg, type_):
      return arg
    return None

  def __init__(self, *args) -> None:
    """Initialize the TypeSig object."""
    self.__type_signature__ = []
    for arg in args:
      if isinstance(arg, type):
        self.__type_signature__.append(arg)
        continue
      e = """The TypeSig must be initialized with types, but received: 
      '%s'!"""
      raise TypeError(monoSpace(e % arg))
    self.setHash(hash((*args,)))

  def __hash__(self, ) -> int:
    """Return the hash of the type signature."""
    if self.__hash_value__ is None:
      e = """The hash value has not been set!"""
      raise ValueError(monoSpace(e))
    return self.__hash_value__

  def setHash(self, hashValue: int) -> None:
    """Set the hash value of the type signature."""
    self.__hash_value__ = hashValue

  def getTypes(self, ) -> list[type]:
    """Return the types of the type signature."""
    return maybe(self.__type_signature__, [])

  def fastCast(self, *args) -> object:
    """Fast cast the arguments to the type signature."""
    if hash((*[type(arg) for arg in args],)) - hash(self):
      return None
    return (*[self.castArg(a, t) for (a, t) in zip(args, self.getTypes())],)

  def cast(self, *args) -> Any:
    """Casts the arguments to the """
    if len(args) != len(self):
      return None
    if not args and not self:
      return []
    out = []
    for (type_, arg) in zip(self.getTypes(), args):
      val = self.castArg(arg, type_)
      if val is None:
        return None
      out.append(val)
    return out

  def __contains__(self, other: tuple) -> bool:
    """Check if the TypeSig is contained in the other tuple."""
    if not other:
      return False
    if isinstance(other, (list, tuple)):
      out = self.cast(*other, )
      return False if out is None else True

  def __bool__(self, ) -> bool:
    """The empty signature makes the instance False."""
    return True if self.__type_signature__ else False

  def __len__(self) -> int:
    """Return the length of the type signature."""
    return len(self.__type_signature__)

  def __str__(self) -> str:
    """String representation"""
    out = [cls.__name__ for cls in self.getTypes()]
    return '(%s,)' % ', '.join(out)

  def __repr__(self) -> str:
    """String representation"""
    return self.__str__()



"""The AbstractNamespace class provides a base class for namespace classes
used in custom metaclasses. The namespace implements a special method
called 'compile' which creates a regular dictionary object from the
namespace. The AbstractMetaclass.__new__ method uses the 'compile' method
on the namespace object to extract the final namespace object. If the
namespace object received does not implement the compile method,
the namespace object is passed to the super call as it is.

AbstractNamespace implements __getitem__, __setitem__ and __delitem__
methods ensuring compliance with the class creation protocol. The
__setitem__ hooks into the __explicit_set__ method, which subclasses may
reimplement. Please note that __getitem__ and __delitem__ provide no such
hooks. Additionally, AbstractNamespace implements getLines which
returns a list of each line in the class body containing the key,
value pair, or key, error pair."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import get_type_hints

from worktoy.parse import maybe
from worktoy.text import typeMsg, stringList

try:
  from typing import Self
except ImportError:
  Self = object

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.meta import Bases


class AbstractNamespace(dict):
  """The AbstractNamespace class provides a base class for namespace classes
  used in custom metaclasses."""

  __meta_class__ = None
  __class_name__ = None
  __base_classes__ = None
  __key_args__ = None
  __class_lines__ = None
  __read_only__ = False
  __iter_contents__ = None

  __first_run__ = True

  __special_keys__ = """__qualname__, __module__, __firstlineno__, 
  __doc__, __annotations__, __dict__, __weakref__, __module__, 
  __metaclass__, __class__, __bases__, __name__, __class_name__,
  __static_attributes__"""

  @classmethod
  def getSpecialKeys(cls) -> list[str]:
    """Getter-function for special keys that are passed to the namespace
    object during class creation from unseen sources. These mysterious
    keys and their values must be respected. The recommended procedure is
    to simply invoke: 'dict.__setitem__(self, key, value)' Not adhering to
    this recommendation may lead to HIGHLY UNDEFINED BEHAVIOUR!"""
    return stringList(cls.__special_keys__)

  @classmethod
  def isSpecialKey(cls, key: str) -> bool:
    """Checks if the key is a special key."""
    return True if key in cls.getSpecialKeys() else False

  def __init__(self, *args, **kwargs) -> None:
    mcls, name, bases = None, None, None
    for arg in args:
      if isinstance(arg, type) and mcls is None:
        mcls = arg
      elif isinstance(arg, str) and name is None:
        name = arg
      elif isinstance(arg, tuple) and bases is None:
        for base in arg:
          if not isinstance(base, type):
            continue
        bases = [*arg, ]
      if all([i is not None for i in [mcls, name, bases]]):
        break
    self.__meta_class__ = mcls
    self.__class_name__ = name
    self.__base_classes__ = bases
    self.__key_args__ = dict(**kwargs)
    dict.__init__(self, )

  def getClassName(self) -> str:
    """Returns the name of the class."""
    return self.__class_name__

  def getMetaClass(self) -> type:
    """Returns the metaclass of the class."""
    return self.__meta_class__

  def getBaseClasses(self) -> Bases:
    """Returns the base classes of the class."""
    return (*self.__base_classes__,)

  def __explicit_set__(self, key: str, value: object) -> None:
    """The __explicit_set__ method is invoked by __setitem__."""

  def getLines(self) -> list[tuple[str, object]]:
    """The getLines method returns a list of each line in the class body
    containing the key, value pair, or key, error pair."""
    clsLines = maybe(self.__class_lines__, [])
    if isinstance(clsLines, list):
      for entry in clsLines:
        if not isinstance(entry, tuple):
          e = typeMsg('entry', entry, tuple)
          raise TypeError(e)
        if entry:
          if not isinstance(entry[0], str):
            e = typeMsg('key', entry[0], str)
            raise TypeError(e)
          if len(entry) != 2:
            e = """Received unexpected number of elements in entry."""
            raise ValueError(e)
      return clsLines
    e = typeMsg('__class_lines__', clsLines, list)
    raise TypeError(e)

  def _appendLine(self, key: str, value: object) -> None:
    """The _appendLine method appends a key, value pair to the class body."""
    existing = self.getLines()
    self.__class_lines__ = [*existing, (key, value)]

  def __getitem__(self, key: str) -> object:
    """First, attempts to retrieve the value from a previous line. """
    try:
      value = dict.__getitem__(self, key)
    except KeyError as keyError:
      self._appendLine(key, keyError)
      raise keyError
    self._appendLine(key, value)
    return value

  def __setitem__(self, key: str, value: object) -> None:
    oldVal = None
    if dict.__contains__(self, key):
      oldVal = dict.__getitem__(self, key)
    self._appendLine(key, (oldVal, value))
    dict.__setitem__(self, key, value)
    self.__explicit_set__(key, value)
    valueType = type(value)
    keySetter = getattr(valueType, 'setNamespaceKey', None)
    if callable(keySetter):
      keySetter(value, key)

  def __delitem__(self, key: str) -> None:
    if dict.__contains__(self, key):
      oldVal = dict.__getitem__(self, key)
      self._appendLine(key, (oldVal, None))
      return dict.__delitem__(self, key)
    e = """Unable to recognize key: '%s'!""" % key
    raise KeyError(e)

  def __iter__(self, ) -> Self:
    """Implementation of the iteration protocol"""
    lines = self.getLines()
    self.__iter_contents__ = [*lines, ]
    return self

  def __next__(self, ) -> object:
    """Implementation of the next method in the iteration protocol"""
    # ic(self.__iter_contents__)
    try:
      return self.__iter_contents__.pop(0)
    except IndexError:
      raise StopIteration

  def compile(self) -> dict:
    """The compile method creates a regular dictionary object from the
    namespace."""
    return {k: v for (k, v) in dict.items(self, )}

  def getAnnotations(self) -> dict[str, type]:
    """Returns a dictionary mapping attribute name to attribute type. This
    is different from the dictionary stored at key '__annotations__',
    which instead stores the name of type only, because of the use of
    annotations from the future import. """
    val = {}
    if dict.__contains__(self, '__annotations__'):
      val = dict.__getitem__(self, '__annotations__')
    return get_type_hints(type('_', (), {'__annotations__': val})())



"""BaseField subclasses AbstractField implementing handling of inheritance
allowing a subclass to reimplement the accessor methods defined in the
parent class. It should be further subclassed to provide an interface for
setting the accessor method keys. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import AbstractField
from worktoy.parse import maybe

try:
  from typing import Callable, Any
except ImportError:
  Callable = object
  Any = object

from worktoy.text import typeMsg, monoSpace


class BaseField(AbstractField):
  """Field provides a flexible implementation of the descriptor
  protocol allowing owning classes to decorate methods as accessor
  methods. """

  __current_owner__ = None
  #  Base accessor methods
  __base_getter__ = None
  __base_setters__ = None
  __base_deleters__ = None
  #  Current accessor methods
  __current_getter__ = None
  __current_setters__ = None
  __current_deleters__ = None

  def _getCurrentOwner(self, **kwargs) -> type:
    """Getter-function for the current owner. """
    if self.__current_owner__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._setCurrentOwner(self.__base_owner__)
      return self._getCurrentOwner(_recursion=True)
    if isinstance(self.__current_owner__, type):
      return self.__current_owner__
    e = typeMsg('__current_owner__', self.__current_owner__, type)
    raise TypeError(e)

  def _setCurrentOwner(self, currentOwner: type) -> None:
    """Setter-function for the current owner. """
    self.__current_owner__ = currentOwner
    self._updateAccessors()

  def _updateAccessors(self, ) -> None:
    """Updates the accessor methods. """
    self._updateGetterFunction()
    self._updateSetterFunctions()
    self._updateDeleterFunctions()

  def _updateGetterFunction(self, ) -> None:
    """Updates the getter method. """
    currentOwner = self._getCurrentOwner()
    getterKey = self._getGetterKey()
    getterFunction = getattr(currentOwner, getterKey, None)
    if getterFunction is None:
      curName = currentOwner.__name__
      fieldName = self.__field_name__
      e = """Current owner: '%s' does not implement a getter method for 
      field: '%s', expected at name: '%s'!"""
      raise AttributeError(monoSpace(e % (curName, fieldName, getterKey)))
    if not callable(getterFunction):
      e = typeMsg('getterFunction', getterFunction, Callable)
      raise TypeError(e)
    self.__current_getter__ = getterFunction

  def _updateSetterFunctions(self, ) -> None:
    """Updates the setter methods. """
    currentOwner = self._getCurrentOwner()
    setterKeys = self._getSetterKeys()
    setterFunctions = []
    for setterKey in setterKeys:
      setterFunction = getattr(currentOwner, setterKey, None)
      if setterFunction is None:
        curName = currentOwner.__name__
        fieldName = self.__field_name__
        e = """Current owner: '%s' does not implement a setter method for 
        field: '%s', expected at name: '%s'!"""
        raise AttributeError(monoSpace(e % (curName, fieldName, setterKey)))
      if not callable(setterFunction):
        e = typeMsg('setterFunction', setterFunction, Callable)
        raise TypeError(e)
      setterFunctions.append(setterFunction)
    self.__current_setters__ = setterFunctions

  def _updateDeleterFunctions(self, ) -> None:
    """Updates the deleter methods. """
    currentOwner = self._getCurrentOwner()
    deleterKeys = self._getDeleterKeys()
    deleterFunctions = []
    for deleterKey in deleterKeys:
      deleterFunction = getattr(currentOwner, deleterKey, None)
      if deleterFunction is None:
        curName = currentOwner.__name__
        fieldName = self.__field_name__
        e = """Current owner: '%s' does not implement a deleter method for 
        field: '%s', expected at name: '%s'!"""
        raise AttributeError(monoSpace(e % (curName, fieldName, deleterKey)))
      if not callable(deleterFunction):
        e = typeMsg('deleterFunction', deleterFunction, Callable)
        raise TypeError(e)
      deleterFunctions.append(deleterFunction)
    self.__current_deleters__ = deleterFunctions

  def _getCurrentGetter(self) -> Callable:
    """Getter-function for the current getter. """
    if self.__current_getter__ is None:
      e = """The getter method is not defined!"""
      raise AttributeError(e)
    if callable(self.__current_getter__):
      return self.__current_getter__
    e = typeMsg('__current_getter__', self.__current_getter__, Callable)
    raise TypeError(e)

  def _getCurrentSetters(self) -> list[Callable]:
    """Getter-function for the current setters. """
    return maybe(self.__current_setters__, [])

  def _getCurrentDeleters(self) -> list[Callable]:
    """Getter-function for the current deleters. """
    return maybe(self.__current_deleters__, [])

  def _validateOwner(self, owner: type, **kwargs) -> type:
    """Validates the owner type. """
    if not isinstance(owner, type):
      e = typeMsg('owner', owner, type)
      raise TypeError(e)
    currentOwner = self._getCurrentOwner()
    if currentOwner is owner:
      return owner
    if kwargs.get('_recursion', False):
      raise RecursionError
    if not issubclass(owner, self.__base_owner__):
      e = """The owner: '%s' is not a subclass of the base owner: '%s'!"""
      thisOwner = self.__base_owner__.__name__
      susOwner = owner.__name__
      raise TypeError(monoSpace(e % (susOwner, thisOwner)))
    self._setCurrentOwner(owner)
    return self._validateOwner(owner, _recursion=True)

  def __get__(self, instance: object, owner: type) -> Any:
    """Getter method for the descriptor. """
    if instance is None:
      return self
    owner = self._validateOwner(owner)
    getter = self._getCurrentGetter()
    return getter(instance)

  def __set__(self, instance: object, value: Any) -> None:
    """Setter method for the descriptor. """
    self._validateOwner(type(instance).__mro__[0])
    setters = self._getCurrentSetters()
    for setter in setters:
      setter(instance, value)

  def __delete__(self, instance: object) -> None:
    """Deleter method for the descriptor. """
    self._validateOwner(type(instance).__mro__[0])
    deleters = self._getCurrentDeleters()
    for deleter in deleters:
      deleter(instance)



"""Dispatcher is a callable class that passes the call to the given
overload that matches the type signature of the arguments received. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.parse import maybe
from worktoy.meta import Overload, TypeSig, DispatchException
from worktoy.text import typeMsg, monoSpace

try:
  from typing import TYPE_CHECKING, Any, Callable
except ImportError:
  TYPE_CHECKING = False
  Any = object
  Callable = object
if TYPE_CHECKING:
  OverloadList = list[Overload]
  SigMap = dict[TypeSig, Overload]
  TypeSigList = list[TypeSig]
else:
  OverloadList = object
  SigMap = object
  TypeSigList = object


class _DispatchCall:
  """This class encapsulates the callable created by the dispatcher. """

  __func__ = None
  __self__ = None
  __name__ = None

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the dispatch call. """
    self.__name__, self.__func__, self.__self__ = [*args, None][:3]

  def __call__(self, *args, **kwargs) -> Any:
    """Calls the function with the given arguments. """
    if self.__self__ is None:
      return self.__func__(*args, **kwargs)
    return self.__func__(self.__self__, *args, **kwargs)


class Dispatcher:
  """Dispatcher is a callable class that passes the call to the given
  overload that matches the type signature of the arguments received. """

  __field_metaclass__ = None
  __field_key__ = None
  __field_name__ = None
  __field_owner__ = None
  __overload_entries__ = None
  __sig_map__ = None
  __class_method_flag__ = None

  def setFieldMetaclass(self, mcls: type) -> None:
    """Sets the field metaclass. """
    if not isinstance(mcls, type):
      e = typeMsg('mcls', mcls, type)
      raise TypeError(e)
    self.__field_metaclass__ = mcls

  def getFieldMetaclass(self, ) -> type:
    """Returns the field metaclass. """
    if self.__field_metaclass__ is None:
      e = """The field metaclass is not set."""
      raise TypeError(e)
    return self.__field_metaclass__

  def setFieldKey(self, fieldKey: str) -> None:
    """Sets the field key. """
    if not isinstance(fieldKey, str):
      e = typeMsg('fieldKey', fieldKey, str)
      raise TypeError(e)
    self.__field_key__ = fieldKey

  def getFieldKey(self, ) -> str:
    """Returns the field key. """
    if self.__field_key__ is None:
      e = """The field key is not set."""
      raise TypeError(e)
    return self.__field_key__

  def setFieldName(self, fieldName: str) -> None:
    """Sets the field name. """
    if not isinstance(fieldName, str):
      e = typeMsg('fieldName', fieldName, str)
      raise TypeError(e)
    self.__field_name__ = fieldName

  def getFieldName(self, ) -> str:
    """Returns the field name. """
    if self.__field_name__ is None:
      e = """The field name is not set."""
      raise TypeError(e)
    return self.__field_name__

  def setFieldOwner(self, fieldOwner: type) -> None:
    """Sets the field owner. """
    mcls = self.getFieldMetaclass()
    if not isinstance(fieldOwner, mcls):
      e = typeMsg('fieldOwner', fieldOwner, mcls)
      raise TypeError(e)
    self.__field_owner__ = fieldOwner

  def getFieldOwner(self, ) -> type:
    """Returns the field owner. """
    if self.__field_owner__ is None:
      e = """The field owner is not set."""
      raise TypeError(e)
    return self.__field_owner__

  def getOverloads(self, ) -> OverloadList:
    """Returns the overload entries. """
    return maybe(self.__overload_entries__, [])

  def getMappings(self, ) -> SigMap:
    """Returns the mappings. """
    if self.__sig_map__ is None:
      e = """The mappings are not set."""
      raise TypeError(e)
    return self.__sig_map__

  #  Main methods

  def getTypeSignatures(self, ) -> TypeSigList:
    """Returns the type signatures of the dispatcher. """
    out = []
    for (sig, fun) in self.getMappings().items():
      out.append(sig)
    return out

  def addOverload(self, overloadEntry: Overload) -> None:
    """Adds an overload entry to the dispatcher. """
    existing = self.getOverloads()
    self.__overload_entries__ = [*existing, overloadEntry]

  def updateMappings(self, ) -> None:
    """Updates the mappings from the new overload settings. """
    cls = self.getFieldOwner()
    mcls = self.getFieldMetaclass()
    self.__sig_map__ = {}
    for entry in self.getOverloads():
      rawSig = entry.getTypeSignature()
      types = []
      for type_ in rawSig:
        if getattr(type_, '__THIS_ZEROTON__', None) is not None:
          types.append(cls)
        elif getattr(type_, '__TYPE_ZEROTON__', None) is not None:
          types.append(mcls)
        elif isinstance(type_, type):
          types.append(type_)
        else:
          e = typeMsg('type', type_, type)
          raise TypeError(e)
      sig = TypeSig(*types, )
      self.__sig_map__[sig] = entry.getWrappedFunction()

  def hereIsMyNumber(self, instance: object) -> Callable:
    """Creates the callable bound to the instance. """

    def callMeMaybe(*args, **kwargs) -> None:
      for (sig, fun) in self.getMappings().items():
        castArg = sig.fastCast(*args, )
        if castArg is None:
          continue
        if instance is None:
          return fun(*args, **kwargs)
        return fun(instance, *args, **kwargs)
      for (sig, fun) in self.getMappings().items():
        castArg = sig.cast(*args, )
        if castArg is None:
          continue
        if instance is None:
          return fun(*args, **kwargs)
        return fun(instance, *args, **kwargs)
      raise DispatchException(self, *args)

    return callMeMaybe

  def __get__(self, instance: object, owner: type) -> Any:
    """Descriptor method for the dispatcher. """
    if instance is None:
      return self
    return self.hereIsMyNumber(instance)

  def __set_name__(self, owner: type, name: str) -> None:
    """This method is invoked when the owning class is created. """
    mcls = self.getFieldMetaclass()
    key = self.getFieldKey()
    if not isinstance(owner, mcls):
      e = """The owner is not derived from the set metaclass!"""
      raise TypeError(e)
    if key != name:
      e = """The key: '%s' does not match the name: '%s'!"""
      raise ValueError(e % (key, name))
    self.setFieldName(name)
    self.setFieldOwner(owner)
    self.updateMappings()

  def __init__(self, fieldMetaclass: type, key: str, **kwargs) -> None:
    """Initializes the dispatcher. """
    self.setFieldMetaclass(fieldMetaclass)
    self.setFieldKey(key)



"""IllegalAttributeException should be raised when a FastObject subclass
tries to define a non-AttriBox attribute."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace


class IllegalAttributeException(Exception):
  """IllegalAttributeException should be raised when a FastObject subclass
  tries to define a non-AttriBox attribute."""

  def __init__(self, key: str, value: object) -> None:
    """The arguments provided should be the key to at which the attribute
    occurs and value should be the value of the attribute."""
    cls = type(value)
    name = cls.__name__
    e = """Attributes are required to be instances of AttriBox, 
    but received '%s' of type '%s' at key: '%s'!""" % (value, name, key)
    Exception.__init__(self, monoSpace(e))



"""FastObject requires all attributes to be instances of AttriBox. This
allows significant performance improvements."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.base import FastMeta


class FastObject(metaclass=FastMeta):
  """FastObject requires all attributes to be instances of AttriBox. This
  allows significant performance improvements."""

  def __init__(self, *args, **kwargs) -> None:
    """LMAO XD!"""

  def __init_subclass__(cls, *args, **kwargs) -> None:
    """LOL this is why we can't have nice things!"""



"""CallMeMeta provides a metaclass for the CallMeMaybe class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import AbstractMetaclass
from worktoy.text import monoSpace

try:
  from typing import Never
except ImportError:
  try:
    from typing import NoReturn as Never
  except ImportError:
    from typing import Any as Never

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False


def _func() -> None:
  """Sample Function"""


_function = type(_func)
_lambda = type(lambda: None)


class _CallMeMeta(AbstractMetaclass):
  """CallMeMeta provides a metaclass for the CallMeMaybe class."""

  def __instancecheck__(cls, callMeMaybe: object) -> bool:
    """Reimplementation recognizing function-like objects. """
    if isinstance(callMeMaybe, (_function, _lambda)):
      return True
    call = getattr(type(callMeMaybe), '__call__', )
    if isinstance(call, (_function, _lambda)):
      return True
    return False

  def __subclasscheck__(cls, subCls: object) -> bool:
    """If 'someObject' pass the '__instancecheck__' method,
    then 'type(someObject)' should pass this method. """

    def inner(arg) -> bool:
      """Inner"""
      if not isinstance(subCls, type):
        return False
      if subCls in [_function, _lambda]:
        return True
      call = getattr(subCls, '__call__', )
      if call in [_function, _lambda]:
        return True
      return False

    out = inner(subCls)
    return out

  def __subclasshook__(cls, __subclass) -> Never:
    """Prevents subclassing of the derived classes. """
    e = """Class '%s' derived from '%s' cannot be subclassed!"""
    raise TypeError(monoSpace(e % (cls.__name__, cls.__class__.__name__)))

  def __call__(cls, *__, **_) -> Never:
    """Prevents the instantiation of the derived classes. """
    e = """Class '%s' derived from '%s' cannot be instantiated!"""
    raise TypeError(monoSpace(e % (cls.__name__, cls.__class__.__name__)))


class CallMeMaybe(metaclass=_CallMeMeta):
  """CallMeMaybe represents types that can be treated as functions. """
  pass


if TYPE_CHECKING:
  from typing import Callable as CallMeMaybe



"""The 'worktoy.desc' implements the descriptor protocol with lazy
instantiation. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._abstract_field import AbstractField
from ._base_field import BaseField

from ._core_descriptor import CoreDescriptor
from ._abstract_descriptor import AbstractDescriptor
from ._bag import Bag
from ._zero_this import THIS
from ._field import Field
from ._attri_box import AttriBox
from ._explicit_box import ExplicitBox



"""The 'worktoy.workWrap' function receives an integer defining character
width and any number of strings. The function then returns a list of
strings containing the words from the strings received such that each
entry in the list does not exceed the character width. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations


def wordWrap(width: int, *textLines) -> list[str]:
  """The wordwrap function wraps the input text to a specified width."""
  words = []
  for line in textLines:
    words.extend(line.split())
  lines = []
  line = []
  while words:
    word = words.pop(0)
    if len(' '.join([*line, word])) <= width:
      line.append(word)
    else:
      lines.append(' '.join(line))
      line = [word]
  if line:
    lines.append(' '.join(line))
  return lines



"""Field class subclasses the 'AbstractDescriptor' and provides the owning
class with decorators for explicitly specifying the accessor methods. This
means that the owning class are free to customize accessor method for each
of their attributes. The GET, SET, DELETE and RESET decorators allow the
owning class to designate the methods responsible for accessor operations.
Besides these decorators, the Field class also inherits the notification
related decorators from the AbstractDescriptor class. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable, Any
except ImportError:
  Callable = object
  Any = object

from worktoy.desc import AbstractDescriptor
from worktoy.text import typeMsg


class Field(AbstractDescriptor):
  """Field provides a flexible implementation of the descriptor
  protocol allowing owning classes to decorate methods as accessor
  methods. """
  __field_type__ = None
  #  Accessor methods
  __getter_function__ = None
  __setter_function__ = None
  __resetter_function__ = None
  __deleter_function__ = None
  #  Keys for accessor methods
  __getter_key__ = None
  __setter_key__ = None
  __resetter_key__ = None
  __deleter_key__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_owner__ = owner
    self.__field_name__ = name
    if self.__getter_key__ is not None:
      self.__getter_function__ = getattr(owner, self.__getter_key__, None)
    if self.__setter_key__ is not None:
      self.__setter_function__ = getattr(owner, self.__setter_key__, None)
    if self.__resetter_key__ is not None:
      self.__resetter_function__ = getattr(
          owner, self.__resetter_key__, None)
    if self.__deleter_key__ is not None:
      self.__deleter_function__ = getattr(owner, self.__deleter_key__, None)

  def getFieldType(self) -> type:
    """Getter-function for the field type."""
    if self.__field_type__ is None:
      return object
    if isinstance(self.__field_type__, type):
      return self.__field_type__
    e = typeMsg('__field_type__', self.__field_type__, type)
    raise TypeError(e)

  def __instance_get__(self, instance: object) -> object:
    """Get the instance object."""
    return self.__get_getter__()(instance)

  def __instance_set__(self, instance: object, value: object) -> None:
    """Set the instance object."""
    self.__get_setter__()(instance, value)

  def __instance_del__(self, instance: object) -> None:
    """Delete the instance object."""
    self.__get_deleter__()(instance)

  def __get_getter__(self, ) -> Callable:
    """Getter-function for the getter-function, getter-ception."""
    if self.__getter_function__ is None:
      e = typeMsg('getter', self.__getter_function__, Callable)
      raise TypeError(e)
    return self.__getter_function__

  def __get_setter__(self, ) -> Callable:
    """Getter-function for the setter-function of the field."""
    if self.__setter_function__ is None:
      e = typeMsg('setter', self.__setter_function__, Callable)
      raise TypeError(e)
    return self.__setter_function__

  def __get_deleter__(self, ) -> Callable:
    """Getter-function for the deleter-function of the field."""
    if self.__deleter_function__ is None:
      e = typeMsg('deleter', self.__deleter_function__, Callable)
      raise TypeError(e)
    return self.__deleter_function__

  def __set_getter__(self, callMeMaybe: Callable) -> Callable:
    """Set the getter function of the field."""
    self.__getter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_setter__(self, callMeMaybe: Callable) -> Callable:
    """Set the setter function of the field."""
    self.__setter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_deleter__(self, callMeMaybe: Callable) -> Callable:
    """Set the deleter function of the field."""
    self.__deleter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def GET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the getter function of the field."""
    return self.__set_getter__(callMeMaybe)

  def SET(self, *args) -> Any:
    """Decorator for setting the setter function of the field."""
    for arg in args:
      if callable(arg):
        return self.__set_setter__(arg)
      if isinstance(arg, str):
        self.__setter_key__ = arg

  def DELETE(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the deleter function of the field."""
    return self.__set_deleter__(callMeMaybe)



"""The 'worktoy.parse' module provides low level parsing and casting
utilities. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._maybe import maybe
from ._maybe_type import maybeType
from ._num_cast_exception import NumCastException
from ._num_cast import numCast, intCast, floatCast, complexCast
from ._type_cast import typeCast



"""AbstractField provides an abstract base class for implementations of
the descriptor protocol that rely on the owning instances to provide the
accessor methods. These methods are identified by keys. One unique key
points to one unique getter method. As for setter and deleter methods,
multiple keys can be assigned to one field. This allows the owning class
to provide multiple setter and deleter methods for one field.

The following methods are provided for specifying keys to accessor methods:
  - setGetterKey: Assigns the getter key. Exactly one getter key should be
    assigned to each field.
  - appendSetterKey: Appends a setter key. Multiple setter keys or none at
    all may be assigned to one field.
  - appendDeleterKey: Appends a deleter key. Multiple deleter keys or none
    at all may be assigned to one field.

Please note that the above methods expects the key string pointing to the
accessor method as input. Thus, they cannot be used as decorators. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.parse import maybe
from worktoy.text import typeMsg


class AbstractField:
  """AbstractField provides an abstract base class for implementations of
  the descriptor protocol that rely on the owning instances to provide the
  accessor methods. This is in contrast to implementations where the
  descriptor classes provide the accessor methods themselves. """

  #  Debug flag
  __print_debug__ = None
  #  Field name and first owner
  __field_name__ = None
  __base_owner__ = None
  #  Accessor method keys
  __getter_key__ = None
  __setter_keys__ = None
  __deleter_keys__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """This method sets the name of the descriptor. """
    self.__field_name__ = name
    self.__base_owner__ = owner

  def _getGetterKey(self, ) -> str:
    """This method returns the getter key. """
    if self.__getter_key__ is None:
      e = """The getter key has not been assigned!"""
      raise AttributeError(e)
    if isinstance(self.__getter_key__, str):
      return self.__getter_key__
    e = typeMsg('__getter_key__', self.__getter_key__, str)
    raise TypeError(e)

  def _getSetterKeys(self, ) -> list[str]:
    """This method returns the setter keys. """
    return maybe(self.__setter_keys__, [])

  def _getDeleterKeys(self, ) -> list[str]:
    """This method returns the deleter keys. """
    return maybe(self.__deleter_keys__, [])

  def setGetterKey(self, setterKey: str) -> None:
    """This method sets the getter key. """
    if self.__getter_key__ is not None:
      e = """The getter key has already been assigned!"""
      raise AttributeError(e)
    if not isinstance(setterKey, str):
      e = typeMsg('setterKey', setterKey, str)
      raise TypeError(e)
    self.__getter_key__ = setterKey

  def appendSetterKey(self, setterKey: str) -> None:
    """This method appends a setter key. """
    existing = self._getSetterKeys()
    self.__setter_keys__ = [*existing, setterKey]

  def appendDeleterKey(self, deleterKey: str) -> None:
    """This method appends a deleter key. """
    existing = self._getDeleterKeys()
    self.__deleter_keys__ = [*existing, deleterKey]

  def __init__(self, *args, **kwargs) -> None:
    """This method initializes the descriptor. """
    if '_root' in kwargs:
      self.__print_debug__ = kwargs['_root']



"""The 'ExplicitBox' class is a subclass of the 'AttriBox' class that
allows the user to explicitly set the default value of the field."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable, Any, Self
except ImportError:
  Callable = object
  Any = object
  Self = object

from worktoy.desc import AttriBox
from worktoy.parse import typeCast


class ExplicitBox(AttriBox):
  """The '_ExplicitBox' class is a subclass of the 'AttriBox' class that
  allows the user to explicitly set the default value of the field."""

  __explicit_default__ = None

  def getExplicitDefault(self) -> Any:
    """This method returns the explicit default value of the field."""
    if self.__explicit_default__ is None:
      e = """The default value of the field is not explicitly set!"""
      raise ValueError(e)
    fieldCls = self.getFieldClass()
    return typeCast(self.__explicit_default__, fieldCls)

  def setExplicitDefault(self, value: Any) -> None:
    """This method sets the explicit default value of the field."""
    if self.__explicit_default__ is not None:
      e = """The default value of the field is already explicitly set!"""
      raise ValueError(e)
    fieldCls = self.getFieldClass()
    self.__explicit_default__ = typeCast(value, fieldCls)

  def __call__(self, *args, **kwargs) -> Self:
    """This method sets the default value of the field."""
    if kwargs:
      e = """The '%s' class dose not allow keyword arguments!"""
      raise TypeError(e % self.__class__.__name__)
    if not args:
      e = """The '%s' class requires one positional argument!"""
      raise TypeError(e % self.__class__.__name__)
    if len(args) - 1:
      e = """The '%s' class requires only one positional argument!"""
      raise TypeError(e % self.__class__.__name__)
    self.setExplicitDefault(args[0])
    return self

  def getDefaultFactory(self) -> Callable:
    """This method returns the default factory function."""
    explicitDefault = self.getExplicitDefault()

    def callMeMaybe(instance: object) -> Any:
      """This function returns the default value of the field."""
      return explicitDefault

    return callMeMaybe



"""MetaNum provides the metaclass for the KeeNum class. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace

try:
  from typing import Self, TYPE_CHECKING, Any
except ImportError:
  TYPE_CHECKING = False
  Self = object
  Any = object

from worktoy.keenum import Num, SpaceNum
from worktoy.meta import AbstractMetaclass
from worktoy.parse import maybe

if TYPE_CHECKING:
  NumList = list[Num]


class MetaNum(AbstractMetaclass):
  """MetaNum provides the metaclass for the KeeNum class. """

  __iter_contents__ = None
  __allow_instantiation__ = None

  @classmethod
  def __prepare__(mcls, name: str, bases: tuple, **kwargs) -> SpaceNum:
    """The __prepare__ method is invoked before the class is created."""
    return SpaceNum(mcls, name, bases, **kwargs)

  def __new__(mcls,
              name: str,
              bases: tuple,
              space: SpaceNum,
              **kwargs) -> type:
    """The __new__ method is invoked when the class is created."""
    namespace = space.compile()
    cls = AbstractMetaclass.__new__(mcls, name, bases, namespace, **kwargs)
    return cls

  def __init__(cls,
               name: str,
               bases: tuple,
               namespace: dict,
               **kwargs) -> None:
    """The __init__ method is invoked when the class is created."""
    AbstractMetaclass.__init__(cls, name, bases, namespace, **kwargs)
    setattr(cls, '__allow_instantiation__', True)
    numEntries = getattr(cls, '__num_entries__', [])
    keenumDict = getattr(cls, '__keenum_dict__', {})
    for num in numEntries:
      name = num.getPublicName()
      publicValue = num.getPublicValue()
      privateValue = num.getPrivateValue()
      keenum = cls(publicValue)
      keenum.setPrivateValue(privateValue)
      keenum.setPublicName(name)
      keenumDict[name] = keenum
    setattr(cls, '__keenum_dict__', keenumDict)
    setattr(cls, '__allow_instantiation__', False)

  def __call__(cls, key: object) -> Any:
    """Get the Num entry by key."""
    allowInstantiation = getattr(cls, '__allow_instantiation__', False)
    if allowInstantiation:
      self = AbstractMetaclass.__call__(cls, key)
      return self
    return cls._resolveNum(key)

  def _getKeeNumList(cls, ) -> list:
    """Get the list of KeeNum instances."""
    keenumDict = getattr(cls, '__keenum_dict__', {})
    keenumList = [v for (k, v) in keenumDict.items()]
    return sorted(keenumList, key=lambda x: int(x))

  def _getKeeNumDict(cls, ) -> dict:
    """Get the dictionary of KeeNum instances."""
    keenumDict = getattr(cls, '__keenum_dict__', {})
    if not keenumDict:
      e = """Class: '%s' has no KeeNum instances!""" % cls.__name__
      raise AttributeError(monoSpace(e))
    return keenumDict

  def _resolveNum(cls, identifier: object) -> Any:
    """Resolve the Num entry."""
    if isinstance(identifier, tuple):
      if len(identifier) == 1:
        return cls._resolveNum(identifier[0])
      e = """Received identifier: '%s' of type: '%s' which is not
      supported!""" % (identifier, type(identifier).__name__)
      raise TypeError(monoSpace(e))
    if isinstance(identifier, cls):
      return identifier
    if isinstance(identifier, str):
      return cls._resolveKey(identifier)
    if isinstance(identifier, int):
      return cls._resolveIndex(identifier)
    e1 = """Received identifier: '%s' of type: '%s' which is not 
    supported!"""
    e2 = """Supported types are: 'str' and 'int'!"""
    actType = type(identifier).__name__
    idStr = str(identifier)
    e = [e1 % (idStr, actType), e2]
    raise TypeError(monoSpace('\n'.join(e)))

  def _resolveKey(cls, key: str, ) -> Any:
    """Resolve a string typed key"""
    keenumDict = cls._getKeeNumDict()
    if key in keenumDict:
      return keenumDict[key]
    e1 = """KeeNum class: '%s' could not resolve the key: '%s'!"""
    e2 = """Supported keys are: \n%s """
    keys = ', '.join([k for k in keenumDict.keys()])
    e = [e1 % (cls.__name__, key), e2 % keys]
    raise KeyError(monoSpace('\n'.join(e)))

  def _rollIndex(cls, index: int) -> int:
    """Roll the index."""
    if index > len(cls):
      e1 = """Index: '%d' exceeds the length of the KeeNum class: '%s' 
      which has only '%d' members!"""
      e2 = e1 % (index, cls.__name__, len(cls))
      raise IndexError(monoSpace(e2))
    if index < 0:
      return cls._rollIndex(len(cls) + index)
    return index

  def _resolveIndex(cls, index: int) -> Any:
    """Resolve an integer typed index."""
    return cls._getKeeNumList()[cls._rollIndex(index)]

  def __iter__(cls, ) -> Self:
    """Implement the iterator protocol."""
    cls.__iter_contents__ = cls._getKeeNumList()
    return cls

  def __next__(cls, ) -> Any:
    """Implement the iterator protocol."""
    if cls.__iter_contents__:
      return cls.__iter_contents__.pop(0)
    raise StopIteration

  def __len__(cls, ) -> int:
    """Implement the length protocol."""
    return len(cls._getKeeNumList())

  def __getitem__(cls, key: object) -> Any:
    """Get the Num entry by key."""
    return cls._resolveNum(key)

  def __contains__(cls, key: object) -> bool:
    """Check if the key is in the KeeNum class."""
    try:
      cls._resolveNum(key)
      return True
    except (KeyError, TypeError):
      return False

  def __getattr__(cls, key: str) -> Any:
    """Get the Num entry by key."""
    try:
      return cls._resolveNum(key)
    except (KeyError, TypeError):
      return object.__getattribute__(cls, key)



"""The 'typeCast' function attempts to cast an object to a given type. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.parse import numCast
from worktoy.text import typeMsg


def typeCast(value: object, target: type) -> object:
  """The 'typeCast' function attempts to cast an object to a given type. """
  if not isinstance(target, type):
    e = typeMsg('target', target, type)
    raise TypeError(e)
  if isinstance(value, target):
    return value
  if target in [int, float, complex]:
    return numCast(value, target)
  try:
    return target(value)
  except Exception as exception:
    e = """Unable to cast object: '%s' of type: '%s' to type: '%s'!"""
    raise TypeError(e % (value, type(value), target)) from exception



"""The 'worktoy.base' package exposes advanced functionality from the
remaining worktoy packages through base classes. The suggested use case is
to inherit from the classes: BaseObject and FastObject.

The BaseObject class expands the standard 'object' base class with support
for function overloading. Any class should be able to inherit from this
class instead of from 'object'.

The FastObject similarly supports function overloading, but restricts the
use of attributes to instances of AttriBox. This allows for significant
performance improvements, whilst retaining the deferred attribute
instantiation provided by AttriBox. Please note however, that FastObject
classes do not in fact own instances of AttriBox after creation. The
metaclass protocol uses the AttriBox instances when creating the classes,
but does not actually pass them on to the created classes.
"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._illegal_attribute_exception import IllegalAttributeException

from ._base_object import BaseObject
from ._fast_space import FastSpace
from ._fast_meta import FastMeta
from ._fast_object import FastObject
from ._overload import overload



"""CoreDescriptor implements the most basic functionalities of the
descriptor protocol without adding any additional features. This includes
in particular the __set_name__ method, which is invoked when the owning
class is created. Then this method informs the descriptor instance of the
class that now owns it and the name by which it appears in the namespace
of the owning class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import BaseMetaclass
from worktoy.text import monoSpace, typeMsg
from re import compile


class CoreDescriptor(metaclass=BaseMetaclass):
  """Implementation of basic descriptor protocol functionalities."""

  __field_name__ = None
  __field_owner__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_owner__ = owner
    self.__field_name__ = name

  def getFieldName(self) -> str:
    """Getter-function for the field name."""
    if self.__field_name__ is None:
      e = """Instance of 'AttriBox' does not belong to class. This 
      typically indicates that the owning class is still being created."""
      raise RuntimeError(monoSpace(e))
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    e = typeMsg('__field_name__', self.__field_name__, str)
    raise TypeError(monoSpace(e))

  def getFieldOwner(self) -> type:
    """Getter-function for the field owner."""
    if self.__field_owner__ is None:
      e = """Instance of 'AttriBox' does not belong to class. This 
      typically indicates that the owning class is still being created. """
      raise RuntimeError(monoSpace(e))
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    e = typeMsg('__field_owner__', self.__field_owner__, type)
    raise TypeError(monoSpace(e))

  def _getPrivateName(self, ) -> str:
    """Returns the name of the private attribute used to store the inner
    object. """
    if self.getFieldName() is None:
      e = """Instance of 'AttriBox' does not belong to class. This 
      typically indicates that the owning class is still being created."""
      raise RuntimeError(monoSpace(e))
    pattern = compile(r'(?<!^)(?=[A-Z])')
    return '__%s__' % pattern.sub('_', self.__field_name__).lower()

  def __init__(self, *args, **kwargs) -> None:
    """Why are we still here?"""

  def __init_subclass__(cls, **kwargs) -> None:
    """Just to suffer?"""



"""BaseNamespace provides the namespace object class for the
BaseMetaclass."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Overload, TypeSig

try:
  from typing import Callable, Any, TYPE_CHECKING
except ImportError:
  Callable = object
  Any = object
  TYPE_CHECKING = False

from worktoy.meta import AbstractNamespace
from worktoy.parse import maybe
from worktoy.text import monoSpace, typeMsg

if TYPE_CHECKING:
  Overloaded = dict[tuple[type, ...], Callable]
  Types = tuple[type, ...]
else:
  Overloaded = object
  Types = object


class OverloadEntry:
  """Instances of this class are used to store the overloads in the
  BaseNamespace."""

  __func_name__ = None
  __call_me_maybe__ = None
  __raw_types__ = None
  __assigned_key__ = None
  __this_zeroton__ = None
  __type_zeroton__ = None

  def __init__(self, *types) -> None:
    self.__raw_types__ = (*types,)

  def getRawTypes(self) -> Types:
    """Getter-function for the raw types"""
    return self.__raw_types__

  def __call__(self, callMeMaybe: Any) -> Callable:
    self.__call_me_maybe__ = callMeMaybe
    self.__func_name__ = callMeMaybe.__name__
    existing = getattr(callMeMaybe, '__type_signatures__', [])
    updated = [*existing, self]
    setattr(callMeMaybe, '__type_signatures__', updated)
    return callMeMaybe

  def assignKey(self, key: str) -> None:
    """Assign the key to the entry."""
    self.__assigned_key__ = key

  def getKey(self, ) -> str:
    """Getter-function for the assigned key."""
    if self.__assigned_key__ is None:
      e = """The key for the overload entry is not assigned!"""
      raise ValueError(monoSpace(e))
    if isinstance(self.__assigned_key__, str):
      return self.__assigned_key__
    e = typeMsg('__assigned_key__', self.__assigned_key__, str, )
    raise TypeError(e)

  def __str__(self, ) -> str:
    """String representation"""
    typeNames = [t.__name__ for t in self.__raw_types__]
    return """%s: %s""" % (self.__func_name__, typeNames)


class BaseNamespace(AbstractNamespace):
  """BaseNamespace provides the namespace object class for the
  BaseMetaclass."""

  __overload_entries__ = None

  def getOverloadEntries(self) -> list[OverloadEntry]:
    """Getter-function for overload entries."""
    return maybe(self.__overload_entries__, [])

  def getOverloadKeys(self) -> list[str]:
    """Getter-function for overload keys."""
    out = []
    for entry in self.getOverloadEntries():
      key = entry.getKey()
      if key not in out:
        out.append(key)
    return out

  def addOverloadEntry(self, entry: OverloadEntry) -> None:
    """Add an overload entry to the namespace."""
    existing = self.getOverloadEntries()
    self.__overload_entries__ = [*existing, entry]

  def getOverload(self, key: str) -> list[OverloadEntry]:
    """Getter-function for an overload entry."""
    out = []
    for entry in self.getOverloadEntries():
      if entry.getKey() == key:
        out.append(entry)
    return out

  def __init__(self, *args, **kwargs) -> None:
    """Initialize the BaseNamespace."""
    AbstractNamespace.__init__(self, *args, **kwargs)

  def __setitem__(self, key: str, value: object) -> None:
    """Set the item in the namespace."""
    if isinstance(value, OverloadEntry):
      value.assignKey(key)
      return self.addOverloadEntry(value)
    if key in self.getOverloadKeys():
      e = """The key '%s' is already assigned to an overload entry!"""
      raise ValueError(monoSpace(e % key))
    return AbstractNamespace.__setitem__(self, key, value)

  def compile(self) -> dict:
    """Compile the namespace into a dictionary."""
    out = {'__overload_entries__': self.getOverloadEntries()}
    return {**AbstractNamespace.compile(self), **out}



"""Num represents enumerated members of a given enum class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import typeMsg

try:
  from typing import Any
except ImportError:
  Any = object


class Num:
  """Num represents enumerated members of a given enum class."""

  __public_value__ = None
  __public_name__ = None
  __private_value__ = None

  def __init__(self, *args, **kwargs) -> None:
    if args:
      self.setPublicValue(args[0])

  def setPrivateValue(self, privateValue: int) -> None:
    """Setter-function for the private value."""
    if self.__private_value__ is not None:
      e = """The private value is already set."""
      raise TypeError(e)
    if not isinstance(privateValue, int):
      e = typeMsg('privateValue', privateValue, int)
      raise TypeError(e)
    self.__private_value__ = privateValue

  def getPrivateValue(self) -> int:
    """Getter-function for the private value."""
    if self.__private_value__ is None:
      e = """The private value is not set."""
      raise TypeError(e)
    if isinstance(self.__private_value__, int):
      return self.__private_value__
    e = typeMsg('privateValue', self.__private_value__, int)
    raise TypeError(e)

  def setPublicName(self, publicName: str) -> None:
    """Setter-function for the public name."""
    if self.__public_name__ is not None:
      e = """The public name is already set."""
      raise TypeError(e)
    if not isinstance(publicName, str):
      e = typeMsg('publicName', publicName, str)
      raise TypeError(e)
    self.__public_name__ = publicName

  def getPublicName(self) -> str:
    """Getter-function for the public name."""
    if self.__public_name__ is None:
      e = """The public name is not set."""
      raise TypeError(e)
    if isinstance(self.__public_name__, str):
      return self.__public_name__
    e = typeMsg('publicName', self.__public_name__, str)
    raise TypeError(e)

  def setPublicValue(self, publicValue: Any) -> None:
    """Setter-function for the public value."""
    if self.__public_value__ is not None:
      e = """The public value is already set."""
      raise TypeError(e)
    self.__public_value__ = publicValue

  def getPublicValue(self, **kwargs) -> Any:
    """Getter-function for the public value."""
    if self.__public_value__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.setPublicValue(self.getPublicName())
      return self.getPublicValue(_recursion=True)
    return self.__public_value__


def auto(*args, **kwargs) -> Num:
  """Auto create a KeeNum class."""
  if args and kwargs:
    return Num((args, kwargs))
  if args:
    if len(args) > 1:
      return Num(args)
    return Num(args[0])
  if kwargs:
    return Num(kwargs)
  return Num()



"""The 'worktoy' package provides a collection of utilities leveraging
advanced python features including custom metaclasses and the descriptor
protocol. The readme file included provides detailed documentation on the
included features. The modules provided depend on each other in
implementation, but can be used independently. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from . import text
from . import parse
from . import meta
from . import desc
from . import base

__all__ = ['desc', 'meta', 'base', 'parse', 'text']



"""Overload encapsulates type signature to function mapping. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.parse import maybe
from worktoy.text import typeMsg

try:
  from typing import TYPE_CHECKING, Callable, Any
except ImportError:
  TYPE_CHECKING = False
  Callable = object
  Any = object

if TYPE_CHECKING:
  RawSigs = list[list[type]]
else:
  RawSigs = object


class Overload:
  """Overload encapsulates type signature to function mapping. """

  __type_signature__ = None
  __wrapped_function__ = None
  __class_method_flag__ = None

  def setTypeSignature(self, *types: type) -> None:
    """Setter-function for the type signature."""
    if self.__type_signature__ is not None:
      e = """The type signature is already set."""
      raise TypeError(e)
    self.__type_signature__ = types

  def getTypeSignature(self, ) -> RawSigs:
    """Getter-function for the type signature."""
    if self.__type_signature__ is None:
      e = """The type signature is not set."""
      raise TypeError(e)
    return self.__type_signature__

  def getWrappedFunction(self, ) -> callable:
    """Getter-function for the wrapped function."""
    if self.__wrapped_function__ is None:
      e = """The wrapped function is not set."""
      raise TypeError(e)
    return self.__wrapped_function__

  def setWrappedFunction(self, wrappedFunction: Callable) -> Callable:
    """Setter-function for the wrapped function."""
    if self.__wrapped_function__ is not None:
      e = """The wrapped function is already set."""
      raise TypeError(e)
    if not callable(wrappedFunction):
      e = """The wrapped function must be callable."""
      raise TypeError(e)
    self.__wrapped_function__ = wrappedFunction
    return wrappedFunction

  def __init__(self, *types: type) -> None:
    self.setTypeSignature(*types)

  def __call__(self, target: Any) -> Any:
    """Decorates target"""
    if isinstance(target, classmethod):
      e = """This version does not implement overload support for 
      classmethods!"""
      raise NotImplementedError(e)
    self.setWrappedFunction(target)
    return self



"""The 'numCast' function receives a numeric value and a numeric type and
returns the value cast as the given type if possible. If unable to cast,
an instance of NumCastException is raised. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.parse import NumCastException

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False


def intCast(value: object) -> int:
  """The 'intCast' function receives a numeric value and returns the value
  cast as an integer if possible. If unable to cast, an instance of
  NumCastException is raised. """
  if isinstance(value, int):
    return value
  if isinstance(value, float):
    if value.is_integer():
      return int(value)
    raise NumCastException(value, int)
  if isinstance(value, complex):
    if not value.imag and value.real.is_integer():
      return int(value.real)
    raise NumCastException(value, int)
  try:
    if TYPE_CHECKING:
      assert isinstance(value, str)  # For the linters out there
    return int(value)
  except Exception as exception:
    raise NumCastException(value, int) from exception


def floatCast(value: object) -> float:
  """The 'floatCast' function receives a numeric value and returns the value
  cast as a float if possible. If unable to cast, an instance of
  NumCastException is raised. """
  if isinstance(value, float):
    return value
  if isinstance(value, int):
    return float(value)
  if isinstance(value, complex):
    if not value.imag:
      return float(value.real)
    raise NumCastException(value, float)
  try:
    if TYPE_CHECKING:
      assert isinstance(value, str)  # For the linters out there
    return float(value)
  except Exception as exception:
    raise NumCastException(value, float) from exception


def complexCast(value: object) -> complex:
  """The 'complexCast' function receives a numeric value and returns the
  value
  cast as a complex number if possible. If unable to cast, an instance of
  NumCastException is raised. """
  if isinstance(value, complex):
    return value
  if isinstance(value, int):
    return floatCast(value) + 0j
  if isinstance(value, float):
    return value + 0j
  try:
    if TYPE_CHECKING:
      assert isinstance(value, str)  # For the linters out there
    return complex(value)
  except Exception as exception:
    raise NumCastException(value, complex) from exception


def numCast(value: object, numType: type) -> object:
  """The 'numCast' function receives a numeric value and a numeric type and
  returns the value cast as the given type if possible. If unable to cast,
  an instance of NumCastException is raised. """

  if numType not in [int, float, complex]:
    try:
      return numType(value)
    except Exception as exception:
      e = """Tried casting object: '%s' to non-standard numeric type: 
      '%s', which raised an exception!"""
      raise TypeError(e % (value, numType,)) from exception
  if numType is int:
    return intCast(value)
  if numType is float:
    return floatCast(value)
  if numType is complex:
    return complexCast(value)



"""OverloadSpace provides the namespace object used to create classes that
support function overloading. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import AbstractNamespace, Dispatcher, Overload
from worktoy.parse import maybe

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

if TYPE_CHECKING:
  DispatchDict = dict[str, Dispatcher]
else:
  DispatchDict = object


class OverloadSpace(AbstractNamespace):
  """OverloadSpace provides the namespace object used to create classes that
  support function overloading. """

  __overloaded_entries__ = None

  def _getOverloadEntries(self) -> dict:
    """Getter-function for the overloaded entries."""
    return maybe(self.__overloaded_entries__, {})

  def appendOverloadEntry(self, key: str, entry: Overload) -> None:
    """Appends an overload entry to the namespace."""
    entries = self._getOverloadEntries()
    existing = entries.get(key, [])
    self.__overloaded_entries__ = {**entries, key: [*existing, entry]}

  def _dispatchFactory(self, ) -> DispatchDict:
    """Build the dispatchers for the overloaded functions."""
    mcls = self.getMetaClass()
    entries = self._getOverloadEntries()
    dispatchers = {}
    bases = self.getBaseClasses()
    dispatcher = None
    for key, overloads in entries.items():
      #  Find dispatcher from base class or create new
      for base in bases:
        dispatcher = getattr(base, key, None)
        if not isinstance(dispatcher, Dispatcher):
          continue
        baseName = base.__name__
        break
      else:
        dispatcher = Dispatcher(mcls, key)
      #  Populate the dispatcher
      for entry in overloads:
        dispatcher.addOverload(entry)
      dispatchers[key] = dispatcher

    return dispatchers

  def compile(self, ) -> dict:
    """Compile the namespace into a dictionary."""
    base = AbstractNamespace.compile(self)
    dispatchers = self._dispatchFactory()
    return {**base, **dispatchers, '__class_body_lines__': self.getLines()}

  def __setitem__(self, key: str, value: object) -> None:
    """Removes instances of OverloadEntry from the default namespace
    handling."""
    if isinstance(value, Overload):
      return self.appendOverloadEntry(key, value)
    if isinstance(value, classmethod):
      e = """Present version does not support """
      return self.__setitem__(key, value.__func__)

    return AbstractNamespace.__setitem__(self, key, value)

  def __getitem__(self, key: str, ) -> object:
    """This reimplementation handles the item retrieval case of
    classmethod. """
    if key == 'classmethod':
      return AbstractNamespace.__getitem__(self, 'derp')
    return AbstractNamespace.__getitem__(self, key)



"""Bag wraps the field object managed by an instance of AttriBox"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from worktoy.meta import BaseMetaclass
from worktoy.parse import typeCast
from worktoy.text import typeMsg


# noinspection PyMissingConstructor
class Bag(metaclass=BaseMetaclass):
  """Bag wraps the field object managed by an instance of AttriBox"""

  __owning_instance__ = None
  __owning_class__ = None
  __inner_object__ = None
  __inner_class__ = None

  def __init__(self, owningInstance: object, innerObject: object) -> None:
    self.setOwningInstance(owningInstance)
    self.setInnerObject(innerObject)

  def getOwningInstance(self) -> object:
    """Getter-function for the owning instance. """
    return self.__owning_instance__

  def setOwningInstance(self, owningInstance: object) -> None:
    """Setter-function for the owning instance. """
    if self.__owning_instance__ is not None:
      if self.__owning_instance__ is owningInstance:
        return
      e = """The owning instance has already been assigned!"""
      raise AttributeError(e)
    self.__owning_instance__ = owningInstance
    self.__owning_class__ = type(owningInstance)

  def getInnerObject(self) -> object:
    """Getter-function for the inner object. """
    return self.__inner_object__

  def setInnerObject(self, innerObject: object) -> None:
    """Setter-function for the inner object. """
    if innerObject is None:
      e = """Attempted to set the inner object to None!"""
      raise RuntimeError(e)
    if self.__inner_class__ is None:
      self.__inner_object__ = innerObject
      self.__inner_class__ = type(innerObject)
    else:
      self.__inner_object__ = typeCast(innerObject, self.getInnerClass())

  def getInnerClass(self) -> type:
    """Getter-function for the inner class. """
    if self.__inner_class__ is None:
      e = """The inner object has not been set!"""
      raise AttributeError(e)
    if isinstance(self.__inner_class__, type):
      return self.__inner_class__
    e = typeMsg('innerClass', self.__inner_class__, type)
    raise TypeError(e)

  def getOwningClass(self) -> type:
    """Getter-function for the owning class. """
    return self.__owning_class__



"""The 'joinWords' function joins a list of words into a single string
with appropriate use of commas and 'and'. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations


def joinWords(*words: str) -> str:
  """Join a list of words into a single string with appropriate use of
  commas and 'and'."""
  if len(words) == 1 and isinstance(words[0], list):
    return joinWords(*words[0])
  if not words:
    return ''
  if len(words) == 1:
    if isinstance(words[0], str):
      return words[0]
    raise TypeError
  if len(words) == 2:
    return '%s and %s' % (words[0], words[1])
  return joinWords(', '.join(words[:-1]), words[-1])



"""AttriBox provides a feature complete implementation of the descriptor
protocol with lazy instantiation. With it, the owning class need only set
one attribute on one line to access the full functionality of the
descriptor. The syntactic sugar is as follows:


class Info:
  # This class provides an attribute class of the Owner class below

  __owning_instance__ = None

  def __init__(self, instance: object) -> None:
    self.__owning_instance__ = instance


class Owner:
  # This class owns attributes through the AttriBox class

  x = AttriBox[float]()
  info = AttriBox[Info](THIS)  # THIS is replaced by the owning instance.


The class of the attribute is placed in the brackets and the parentheses
are given the arguments used to instantiate the attribute. It is possible
to pass special placeholders here which are replaced when the attribute
object is created. The placeholders are:

THIS: The owning instance
TYPE: The owning class
BOX: The AttriBox instance
ATTR: The attribute class or its subclass

The lifecycle of the AttriBox instance is as follows:

1. The AttriBox class itself is created
2. The AttriBox instance is created during the class body execution of a
class that is being created.
3. When the class creation process completes, the '__set_name__' method is
invoked. This class is inherited from the 'CoreDescriptor' class.
4. When this AttriBox instance is accessed through the owning class,
not an instance of it, the AttriBox instance itself returns.
5. When the access is through an instance of the owning class,
the AttriBox instance first attempts to find an existing value in the
namespace of the instance at its private name. This value is returned if
it exists.
6. Otherwise, an instance of the wrapping class 'Bag' is created and an
instance of the inner class is created and stored in the 'Bag' instance.
It is the 'Bag' instance that is stored in the namespace of the owning
class and during calls to __get__, the wrapped object is returned from
inside the Bag instance.
7. By default, the setter method expects the value received to be of the
same type as the existing object in the Bag instance.
8. By default, the deleter method is disabled and will raise an exception.
This is because calls on the form: 'del instance.attribute' must then
cause 'instance.attribute' to result in Attribute error. This is not
really practical as it is insufficient to remove the value as the
descriptor is on the owning class not the instance. This means that no
functionality is present to distinguish between the __delete__ having been
called, and then inner object not having been created yet."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Self, Callable
except ImportError:
  Self = object
  Callable = object

try:
  from typing import Never
except ImportError:
  Never = object

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

from typing import Any
from worktoy.desc import THIS
from worktoy.desc import AbstractDescriptor, Bag
from worktoy.parse import maybe, typeCast
from worktoy.text import typeMsg, monoSpace


class AttriBox(AbstractDescriptor):
  """AttriBox class improves the AttriBox class!"""

  __default_object__ = None
  __field_class__ = None
  __pos_args__ = None
  __key_args__ = None

  @staticmethod
  def getThisFilter(instance: object) -> Callable:
    """Getter-function for the 'THIS' filter."""

    def thisFilter(obj: object) -> object:
      """Filter for the 'THIS' placeholder."""
      if obj is THIS:
        return instance
      return obj

    return thisFilter

  @classmethod
  def __class_getitem__(cls, fieldClass: type) -> AttriBox:
    """Class method for creating a AttriBox instance."""
    return cls(fieldClass)

  def __init__(self, *args) -> None:
    AbstractDescriptor.__init__(self)
    for arg in args:
      if isinstance(arg, type):
        fieldClass = arg
        break
    else:
      e = """AttriBox constructor requires the fieldClass type!"""
      raise ValueError(e)
    if isinstance(fieldClass, type):
      self.__field_class__ = fieldClass
    else:
      e = """AttriBox constructor requires the fieldClass type!"""
      e2 = typeMsg('fieldClass', fieldClass, type)
      raise TypeError(monoSpace('%s\n%s' % (e, e2)))
    self.__field_class__ = fieldClass

  def __call__(self, *args, **kwargs) -> Self:
    self.__pos_args__ = [*args, ]
    self.__key_args__ = {**kwargs, }
    return self

  def getFieldClass(self, ) -> type:
    """Getter-function for the field class."""
    if self.__field_class__ is None:
      e = """The field class of the AttriBox instance has not been set!"""
      raise AttributeError(e)
    if isinstance(self.__field_class__, type):
      return self.__field_class__
    e = typeMsg('__field_class__', self.__field_class__, type)
    raise TypeError(e)

  def getArgs(self, instance: object, **kwargs) -> list[Any]:
    """Getter-function for positional arguments"""
    if kwargs.get('_root', False):
      return maybe(self.__pos_args__, [])
    thisFilter = self.getThisFilter(instance)
    return [thisFilter(arg) for arg in maybe(self.__pos_args__, [])]

  def getKwargs(self, instance: object, **kwargs) -> dict[str, Any]:
    """Getter-function for keyword arguments"""
    if kwargs.get('_root', False):
      return maybe(self.__key_args__, {})
    thisFilter = self.getThisFilter(instance)
    kw = maybe(self.__key_args__, {})
    return {k: thisFilter(v) for (k, v) in kw.items()}

  def getDefaultFactory(self) -> Any:
    """Getter-function for function creating the default value. """
    keyArgs = self.getKwargs(None, _root=True)
    posArgs = self.getArgs(None, _root=True)
    fieldClass = self.getFieldClass()

    def callMeMaybe(instance: object) -> Any:
      """This function creates the default value."""
      thisFilter = self.getThisFilter(instance)
      newArgs = [thisFilter(arg) for arg in posArgs]
      newKeys = {k: thisFilter(v) for (k, v) in keyArgs.items()}

      if fieldClass is bool:
        innerObject = True if [*newArgs, None][0] else False
      else:
        if newArgs and newKeys:
          innerObject = fieldClass(*newArgs, **newKeys)
        elif newArgs:
          innerObject = fieldClass(*newArgs)
        elif newKeys:
          innerObject = fieldClass(**newKeys)
        else:
          innerObject = fieldClass()
      if TYPE_CHECKING:
        return Bag(None, innerObject)
      return innerObject

    return callMeMaybe

  def createFieldObject(self, instance: object, ) -> Bag:
    """Create the field object. If the default object is set, it is used."""
    factory = self.getDefaultFactory()
    return Bag(instance, factory(instance))

  def __instance_reset__(self, instance: object) -> None:
    """Reset-function for the instance."""
    pvtName = self._getPrivateName()
    delattr(instance, pvtName)

  def __instance_get__(self, instance: object, **kwargs) -> Any:
    """Getter-function for the instance. Please note, that if the call is
    the notifier asking what the previous value was, the functionality in
    the AbstractDescriptor expects and handles the exception. """
    pvtName = self._getPrivateName()
    bag = getattr(instance, pvtName, None)
    if bag is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      innerObject = self.createFieldObject(instance)
      setattr(instance, pvtName, innerObject)
      return self.__instance_get__(instance, _recursion=True)
    if not isinstance(bag, Bag):
      e = typeMsg('bag', bag, Bag)
      raise TypeError(e)
    innerObject = bag.getInnerObject()
    fieldClass = self.getFieldClass()
    if isinstance(innerObject, fieldClass):
      return innerObject
    e = typeMsg('innerObject', innerObject, fieldClass)
    raise TypeError(e)

  def __instance_set__(self, instance: Any, value: Any, **kwargs) -> None:
    """Setter-function for the instance."""
    pvtName = self._getPrivateName()
    fieldCls = self.getFieldClass()
    value = typeCast(value, fieldCls)
    if value is None:
      e = """The '%s' object received a 'None' value in the setter. This 
      object has field name: '%s', field class: '%s' and owner name: '%s'!"""
      boxName = type(self).__name__
      fieldName = self.getFieldName()
      fieldCls = self.getFieldClass().__name__
      ownerName = type(instance).__name__
      e2 = monoSpace(e % (boxName, fieldName, fieldCls, ownerName))
      raise ValueError(e2)
    bag = getattr(instance, pvtName, None)
    if bag is None:
      return setattr(instance, pvtName, Bag(instance, value))
    bag.setInnerObject(value)
    return setattr(instance, pvtName, bag)

  if TYPE_CHECKING:
    from typing import NoReturn as Never

  def __instance_del__(self, instance: object) -> Never:
    """Deleter-function for the instance."""
    e = """Deleter-function is not implemented by the AttriBox class."""
    raise TypeError(e)

  def __str__(self, ) -> str:
    """String representation"""
    posArgs = self.getArgs(None, _root=True)
    keyArgs = self.getKwargs(None, _root=True)
    posStr = ', '.join([str(arg) for arg in posArgs])
    keyStr = ', '.join([f'{k}={v}' for (k, v) in keyArgs.items()])
    argStr = ', '.join([arg for arg in [posStr, keyStr] if arg])
    clsName = self.getFieldClass().__name__
    return """AttriBox[%s](%s)""" % (clsName, argStr)

  def __repr__(self, ) -> str:
    """String representation"""
    return str(self)



"""Zeroton introduces the novel idea of a class not having even one
instance, unlike a singleton class. The purpose of such classes is to have
universal objects that retain identity across modules."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

try:
  from typing import Callable
except ImportError:
  Callable = object

from worktoy.meta import BaseMetaclass, AbstractNamespace
from worktoy.text import typeMsg


class ZeroSpace(AbstractNamespace):
  """The Zeroton class is a class that has no instances. It is used to
  represent a universal object that retains identity across modules."""

  __virgin_namespace__ = None

  def setVirginNamespace(self, namespace: AbstractNamespace) -> None:
    """This method sets the virgin namespace."""
    self.__virgin_namespace__ = namespace

  def getVirginNamespace(self) -> AbstractNamespace:
    """This method returns the virgin namespace."""
    return self.__virgin_namespace__

  def __getitem__(self, key: str) -> Any:
    """If there is already a virgin namespace, this class always raises a
    KeyError"""
    if self.__virgin_namespace__ is None:
      return AbstractNamespace.__getitem__(self, key)
    raise KeyError(key)

  def __setitem__(self, key: str, value: object) -> None:
    """If there is already a virgin namespace, this class always raises a
    KeyError"""
    if self.__virgin_namespace__ is None:
      AbstractNamespace.__setitem__(self, key, value)


class ZeroMeta(BaseMetaclass):
  """The Zeroton class is a class that has no instances. It is used to
  represent a universal object that retains identity across modules."""

  __all__ = {}

  def __instancecheck__(cls, instance: Any) -> bool:
    """I know this is a bit lame, but the documentation took a very long
    time. You're welcome!"""
    if cls.__name__ == 'ATTR':
      if instance.__name__ == 'BOX':
        return True
      return False
    if cls.__name__ == 'TYPE':
      if instance.__name__ == 'THIS':
        return True
    return False

  @classmethod
  def __prepare__(mcls, name: str, *args, **kwargs) -> ZeroSpace:
    if mcls.__all__.get(name, None) is not None:
      return mcls.__all__.get(name).get('space')
    return ZeroSpace(mcls, name, *args, **kwargs)

  def __new__(mcls, name: str, *args, **kwargs) -> type:
    if mcls.__all__.get(name, None) is not None:
      return mcls.__all__.get(name).get('cls')
    return BaseMetaclass.__new__(ZeroMeta, name, *args, **kwargs)

  def __call__(cls, *args, **kwargs):
    clsCall = getattr(cls, '__class_call__', None)
    if clsCall is not None:
      if callable(clsCall):
        if getattr(clsCall, '__self__', None) is None:
          clsCall(cls, *args, **kwargs)
        else:
          clsCall(*args, **kwargs)
      else:
        e = typeMsg('clsCall', clsCall, Callable)
        raise TypeError(e)
    return cls


class Zeroton(metaclass=ZeroMeta):
  """The Zeroton class is a class that has no instances. It is used to
  represent a universal object that retains identity across modules."""



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. Because the overload protocol relies on type matching,
this exception subclasses TypeError such that it can be caught by external
error handlers. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.meta import Dispatcher


class DispatchException(TypeError):
  """DispatchException provides a custom exception raised when an instance
  of OverloadDispatcher fails to resolve the correct function from the
  given arguments. """

  def __init__(self, dispatch: Dispatcher, *args) -> None:
    e = """Dispatch instance could not match signature of arguments to any 
    supported signature. Received: <br><tab>'%s'<br>Supported 
    signatures: %s"""
    argSig = ', '.join([type(arg).__name__ for arg in args])
    supported = []
    for sig in dispatch.getTypeSignatures():
      types = sig.getTypes()
      typeNames = [t.__name__ for t in types]
      supported.append("""(%s)""" % ', '.join(typeNames))
    supportedSig = '<br><tab>'.join(supported)
    e2 = monoSpace(e % (argSig, supportedSig))
    TypeError.__init__(self, e2)



"""NumCastException is a custom exception subclass of ValueError raised by
the 'numCast' function when unable to caste a value to a given type. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations


class NumCastException(TypeError):
  """NumCastException is a custom exception subclass of ValueError raised by
  the 'numCast' function when unable to caste a value to a given type. """

  def __init__(self, value: object, target: type) -> None:
    e = """Unable to cast object: '%s' to numeric type: '%s'!"""
    TypeError.__init__(self, e % (value, target))



"""The 'worktoy.keenum' package provides the KeeNum class for enumeration!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._num import Num, auto
from ._space_num import SpaceNum
from ._meta_num import MetaNum
from ._keenum import KeeNum



"""Field class subclasses the 'AbstractDescriptor' and provides the owning
class with decorators for explicitly specifying the accessor methods. This
means that the owning class are free to customize accessor method for each
of their attributes. The GET, SET, DELETE and RESET decorators allow the
owning class to designate the methods responsible for accessor operations.
Besides these decorators, the Field class also inherits the notification
related decorators from the AbstractDescriptor class. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import BaseField
from worktoy.parse import maybe

try:
  from typing import Callable, Any, TYPE_CHECKING
except ImportError:
  Callable = object
  Any = object
  TYPE_CHECKING = False

from worktoy.text import typeMsg, monoSpace

if TYPE_CHECKING:
  from worktoy.meta import CallMeMaybe


class Field(BaseField):
  """Field provides a flexible implementation of the descriptor
  protocol allowing owning classes to decorate methods as accessor
  methods. """

  def GET(self, callMeMaybe: CallMeMaybe) -> CallMeMaybe:
    """Decorator for the getter-method."""
    self.setGetterKey(callMeMaybe.__name__)
    return callMeMaybe

  def SET(self, callMeMaybe: CallMeMaybe) -> CallMeMaybe:
    """Decorator for the setter-method."""
    self.appendSetterKey(callMeMaybe.__name__)
    return callMeMaybe

  def DELETE(self, callMeMaybe: CallMeMaybe) -> CallMeMaybe:
    """Decorator for the deleter-method."""
    self.appendDeleterKey(callMeMaybe.__name__)
    return callMeMaybe



"""KeeNum enumerates items."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import Field
from worktoy.keenum import MetaNum
from worktoy.meta import CallMeMaybe
from worktoy.text import typeMsg


def _root(callMeMaybe: CallMeMaybe) -> CallMeMaybe:
  """Return the root class."""
  setattr(callMeMaybe, '__trust_me_bro__', True)
  return callMeMaybe


class KeeNum(metaclass=MetaNum):
  """KeeNum enumerates items."""

  __private_value__ = None
  __public_value__ = None
  __public_name__ = None

  value = Field()
  name = Field()

  @value.GET
  def _getPublicValue(self) -> object:
    """Getter-function for the public value."""
    if self.__public_value__ is None:
      return self.name
    return self.__public_value__

  @name.GET
  def _getPublicName(self) -> str:
    """Getter-function for the public name."""
    if self.__public_name__ is None:
      e = """The public name is not set."""
      raise TypeError(e)
    if isinstance(self.__public_name__, str):
      return self.__public_name__
    e = typeMsg('publicName', self.__public_name__, str)
    raise TypeError(e)

  def setPrivateValue(self, privateValue: int) -> None:
    """Setter-function for the private value."""
    if self.__private_value__ is not None:
      e = """The private value is already set."""
      raise TypeError(e)
    if not isinstance(privateValue, int):
      e = typeMsg('privateValue', privateValue, int)
      raise TypeError(e)
    self.__private_value__ = privateValue

  def setPublicName(self, publicName: str) -> None:
    """Setter-function for the public name."""
    if self.__public_name__ is not None:
      e = """The public name is already set."""
      raise TypeError(e)
    if not isinstance(publicName, str):
      e = typeMsg('publicName', publicName, str)
      raise TypeError(e)
    self.__public_name__ = publicName

  def __int__(self, ) -> int:
    """Return the private value as an integer."""
    if self.__private_value__ is None:
      e = """The private value is not set."""
      raise TypeError(e)
    if isinstance(self.__private_value__, int):
      return self.__private_value__
    e = typeMsg('privateValue', self.__private_value__, int)
    raise TypeError(e)

  @_root
  def __init__(self, publicValue: object) -> None:
    self.__public_value__ = publicValue

  def __str__(self, ) -> str:
    """Return the public name as a string."""
    clsName = type(self).__name__
    return """%s.%s""" % (clsName, self.name.upper())

  def __repr__(self, ) -> str:
    """Return the public name as a string."""
    clsName = type(self).__name__
    return """%s(%s)""" % (clsName, self.name)

  def __eq__(self, other: object) -> bool:
    """Return True if the public value is equal to the other."""
    if isinstance(other, KeeNum):
      cls = type(self).__mro__[0]
      otherCls = type(other).__mro__[0]
      if cls is otherCls:
        return False if int(self) - int(other) else True
      return False
    if isinstance(other, int):
      return False if int(self) - other else True
    if isinstance(other, str):
      return True if self.name == other else False
    return False

  def __hash__(self, ) -> int:
    """Return the hash of the public value."""
    return hash((type(self), self.name, int(self)))



"""FastSpace provides the namespace object class for the FastObject class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import typeMsg

try:
  from typing import Callable, Any, Self
except ImportError:
  Callable = object
  Any = object
  Self = object

from worktoy.desc import AttriBox, ExplicitBox
from worktoy.meta import OverloadSpace, Overload, CallMeMaybe
from worktoy.parse import maybe


class FastSpace(OverloadSpace):
  """EZSpace provides the namespace object class for the EZData class."""

  __field_boxes__ = None
  __inner_functions__ = None

  def _getFieldBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the field boxes."""
    return maybe(self.__field_boxes__, [])

  def _getBaseBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the base field boxes."""
    baseClasses = self.getBaseClasses()
    if not baseClasses:
      return []
    baseCls = self.getBaseClasses()[0]
    baseBoxes = getattr(baseCls, '__field_boxes__', None)
    if baseBoxes is None:
      return []
    if isinstance(baseBoxes, list):
      return baseBoxes
    e = typeMsg('baseBoxes', baseBoxes, list)
    raise TypeError(e)

  def _getAllBoxes(self, ) -> list[tuple[str, AttriBox]]:
    """This method returns all the field boxes."""
    baseBoxes = self._getBaseBoxes()
    fieldBoxes = self._getFieldBoxes()
    return [*baseBoxes, *fieldBoxes]

  def _addFieldBox(self, key: str, box: AttriBox) -> None:
    """This method adds a field box to the namespace."""
    boxes = self._getFieldBoxes()
    self.__field_boxes__ = [*boxes, (key, box)]

  def __setitem__(self, key: str, value: object) -> None:
    """This method sets the key, value pair in the namespace."""
    if isinstance(value, AttriBox):
      return self._addFieldBox(key, value)
    if callable(value) or isinstance(value, Overload):
      return OverloadSpace.__setitem__(self, key, value)
    if self.isSpecialKey(key):
      return dict.__setitem__(self, key, value)
    return self.__setitem__(key, ExplicitBox[type(value)](value))

  @staticmethod
  def _getattrFactory(boxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__getattr__' method which automatically
    retrieves the AttriBox instances."""

    keys = [key for (key, box) in boxes]
    defGet = {key: box.getDefaultFactory() for (key, box) in boxes}

    def __getattr__(self, key: str) -> object:
      """This automatically generated '__getattr__' method retrieves the
      AttriBox instances."""
      if key in defGet:
        setattr(self, key, defGet[key](self))
      return object.__getattribute__(self, key)

    return __getattr__

  @staticmethod
  def _initFactory(attriBoxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__init__' method which automatically
    populates the AttriBox instances."""

    keys = [key for (key, box) in attriBoxes]
    defVals = {key: box.getDefaultFactory() for (key, box) in attriBoxes}
    valTypes = {key: box.getFieldClass() for (key, box) in attriBoxes}

    def __init__(self, *args, **kwargs) -> None:
      """This automatically generated '__init__' method populates the
      AttriBox instances."""
      initValues = dict()  # Creates temporary dictionary for initial values

      #  Retrieves values from positional arguments
      for (i, (key, type_)) in enumerate(valTypes.items()):
        if len(args) > i:
          if isinstance(args[i], type_):
            initValues[key] = args[i]
          else:
            e = typeMsg(key, args[i], type_)
            raise TypeError(e)

      #  Retrieves values from keyword arguments
      for (key, type_) in valTypes.items():
        if key in kwargs:
          val = kwargs[key]
          if isinstance(val, type_):
            initValues[key] = val
          else:
            e = typeMsg(key, val, type_)
            raise TypeError(e)

      #  Creates default values for missing arguments
      for (key, defVal) in defVals.items():
        if key not in initValues:
          initValues[key] = defVal(self)

      #  Assigns values
      for (key, value) in initValues.items():
        if value is None:
          setattr(self, key, value)

    return __init__

  @staticmethod
  def _slotsFactory(boxes: list[tuple[str, AttriBox]]) -> list[str]:
    """This factory creates the '__slots__' list which is used to restrict
    the namespace to the AttriBox instances."""
    return [box[0] for box in boxes]

  def compile(self) -> dict:
    """The namespace created by the BaseNamespace class is updated with
    the '__init__' function created by the factory function."""
    namespace = OverloadSpace.compile(self)
    oldInit = namespace.get('__init__', None)
    oldGetAttr = namespace.get('__getattr__', None)
    boxes = self._getAllBoxes()
    namespace['__field_boxes__'] = self._getFieldBoxes()
    namespace['__slots__'] = self._slotsFactory(boxes)
    if oldInit is None or oldInit is object.__init__:
      newInit = self._initFactory(boxes)
      namespace['__init__'] = newInit
    if oldGetAttr is None:
      newGetAttr = self._getattrFactory(boxes)
      namespace['__getattr__'] = newGetAttr
    return namespace



"""AbstractDescriptor subclasses the CoreDescriptor and expands the basic
descriptor functionality with the following functionalities:

 - Notification hooks for each accessor
 - Instance specific silencing of notifications
 - Global suppression of notifications by the descriptor instance
 - The basic accessor methods '__get__', '__set__' and '__delete__' are
 replaced with explicit methods, except for '__get__' which returns the
 descriptor instance itself, when the owning instance passed to it is
 None. This is the case when the descriptor is accessed on the owning
 class directly. This distinction is important, as it is the only way to
 reach the actual descriptor instance.
  - Subclasses should implement the instance specific accessor functions
  as appropriate for their intended function. Only the '__instance_get__'
  is strictly required. The other methods will raise a TypeError if
  invoked. Subclasses are free to reimplement this.
  - This class implements the notification mechanisms leaving subclasses
  with the above instance specific accessors.
  - The notification system requires that owning class should decorate the
  methods it wishes to be notified of with the ONGET, ONSET and ONDEL
  decorators.

"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Any, Callable

from worktoy.desc import CoreDescriptor
from worktoy.parse import maybe
from worktoy.text import monoSpace

try:
  from typing import Never
except ImportError:
  Never = object


class AbstractDescriptor(CoreDescriptor):
  """AbstractDescriptor provides common boilerplate for descriptor
  classes. """
  __on_set_callbacks__ = None
  __pre_set_callbacks__ = None
  __on_del_callbacks__ = None
  __pre_del_callbacks__ = None
  __pre_get_callbacks__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Wrapper factories

  @staticmethod
  def wrapGet(callMeMaybe: Callable) -> Callable:
    """This method creates a flexible wrapper for the using the callable
    as a getter. """

    def wrapper(instance: object, value: object) -> Any:
      """Wrapper function handling the notification. The wrapper first
      tries passing the instance and values to the given callable. If the
      callable raises a TypeError relating to 'positional arguments', it
      will then attempt to call with only the instance. """
      try:
        return callMeMaybe(instance, value)
      except TypeError as typeError:
        if 'positional arguments' in str(typeError):
          try:
            return callMeMaybe(value)
          except Exception as exception:
            raise exception from typeError
        else:
          raise typeError

    return wrapper

  @staticmethod
  def wrapDel(callMeMaybe: Callable) -> Callable:
    """Wrapper function for the deleter notifier."""

    def wrapper(instance: object, value: object) -> None:
      """Wrapper function handling the notification. The wrapper first
      tries passing the instance and values to the given callable. If the
      callable raises a TypeError relating to 'positional arguments', it
      will then attempt to call with only the instance. """
      try:
        callMeMaybe(instance, value)
      except TypeError as typeError:
        if 'positional arguments' in str(typeError):
          try:
            callMeMaybe(value)
          except Exception as exception:
            raise exception from typeError
        else:
          raise typeError

    return wrapper

  @staticmethod
  def wrapSet(callMeMaybe: Callable) -> Callable:
    """This method creates a flexible wrapper for the using the callable
    as a setter. """

    def wrapper(instance: object, oldVal: object, newVal: object) -> None:
      """Wrapper function handling the notification. The wrapper first
      tries passing the instance, old value and new value to the given
      callable. If the callable raises a TypeError relating to 'positional
      arguments', it will then attempt to call with only the instance and
      the new value. """
      try:
        callMeMaybe(instance, oldVal, newVal)
      except TypeError as typeError:
        if 'positional arguments' in str(typeError):
          try:
            callMeMaybe(instance, newVal)
          except Exception as exception:
            raise exception from typeError
        else:
          raise typeError

    return wrapper

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Getter-functions for notification callbacks

  def _onSetCallbacks(self, ) -> tuple[Callable, ...]:
    """This method returns the set callbacks."""
    return maybe(self.__on_set_callbacks__, [])

  def _preSetCallbacks(self, ) -> tuple[Callable, ...]:
    """This method returns the set callbacks."""
    return maybe(self.__pre_set_callbacks__, [])

  def _onDelCallbacks(self, ) -> tuple[Callable, ...]:
    """This method returns the set callbacks."""
    return maybe(self.__on_del_callbacks__, [])

  def _preDelCallbacks(self, ) -> tuple[Callable, ...]:
    """This method returns the set callbacks."""
    return maybe(self.__pre_del_callbacks__, [])

  def _preGetCallbacks(self, ) -> tuple[Callable, ...]:
    """This method returns the set callbacks."""
    return maybe(self.__pre_get_callbacks__, [])

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Append-functions for notification callbacks

  def _addPreSetCallback(self, callMeMaybe: Callable) -> None:
    """Registers the given callable to receive notifications when the
    value is about to change. """

    preSetCallbacks = self._preSetCallbacks()
    wrapper = self.wrapSet(callMeMaybe)
    self.__pre_set_callbacks__ = [*preSetCallbacks, wrapper]

  def _addOnSetCallback(self, callMeMaybe: Callable) -> None:
    """Registers the given callable to receive notifications when the
    value has changed. """

    onSetCallbacks = self._onSetCallbacks()
    wrapper = self.wrapSet(callMeMaybe)
    self.__on_set_callbacks__ = [*onSetCallbacks, wrapper]

  def _addPreDelCallback(self, callMeMaybe: Callable) -> None:
    """Registers the given callable to receive notifications when the
    value is about to be deleted. """

    preDelCallbacks = self._preDelCallbacks()
    wrapper = self.wrapDel(callMeMaybe)
    self.__pre_del_callbacks__ = [*preDelCallbacks, wrapper]

  def _addOnDelCallback(self, callMeMaybe: Callable) -> None:
    """Registers the given callable to receive notifications when the
    value has been deleted. """

    onDelCallbacks = self._onDelCallbacks()
    wrapper = self.wrapDel(callMeMaybe)
    self.__on_del_callbacks__ = [*onDelCallbacks, wrapper]

  def _addPreGetCallback(self, callMeMaybe: Callable) -> None:
    """Registers the given callable to receive notifications when the
    value is about to be accessed. """

    preGetCallbacks = self._preGetCallbacks()
    wrapper = self.wrapGet(callMeMaybe)
    self.__pre_get_callbacks__ = [*preGetCallbacks, wrapper]

  @staticmethod
  def _addOnGetCallback(*_) -> Never:
    """It is not possible to be notified 'after' the getter has returned!"""
    e = """It is not possible to be notified 'after' the getter has 
    returned!"""
    raise TypeError(monoSpace(e))

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Decorators for subscribing to the notifications

  @staticmethod
  def ONGET(*_) -> Never:
    """It is not possible to be notified 'after' the getter has returned!"""
    e = """It is not possible to be notified 'after' the getter has 
    returned!"""
    raise TypeError(monoSpace(e))

  def ONSET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for subscribing to the setter."""
    self._addOnSetCallback(callMeMaybe)
    return callMeMaybe

  def ONDEL(self, callMeMaybe: Callable) -> Callable:
    """Decorator for subscribing to the deleter."""
    self._addOnDelCallback(callMeMaybe)
    return callMeMaybe

  def PREGET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for subscribing to the pre-getter."""
    self._addPreGetCallback(callMeMaybe)
    return callMeMaybe

  def PRESET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for subscribing to the pre-setter."""
    self._addPreSetCallback(callMeMaybe)
    return callMeMaybe

  def PREDEL(self, callMeMaybe: Callable) -> Callable:
    """Decorator for subscribing to the pre-deleter."""
    self._addPreDelCallback(callMeMaybe)
    return callMeMaybe

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Dispatcher methods for the notifications

  @staticmethod
  def notifyOnGET(*_) -> Never:
    """It is not possible to be notified 'after' the getter has returned!"""
    e = """It is not possible to be notified 'after' the getter has 
    returned!"""
    raise TypeError(monoSpace(e))

  def notifyPreGet(self, instance: object, *values) -> None:
    """Dispatches the pre-get notifications."""
    for callMeMaybe in self._preGetCallbacks():
      callMeMaybe(instance, *values)

  def notifyOnSet(self, instance: object, *values) -> None:
    """Dispatches the on-set notifications."""
    for callMeMaybe in self._onSetCallbacks():
      callMeMaybe(instance, *values)

  def notifyPreSet(self, instance: object, *values) -> None:
    """Dispatches the pre-set notifications."""
    for callMeMaybe in self._preSetCallbacks():
      callMeMaybe(instance, *values)

  def notifyOnDel(self, instance: object, *values) -> None:
    """Dispatches the on-del notifications."""
    for callMeMaybe in self._onDelCallbacks():
      callMeMaybe(instance, *values)

  def notifyPreDel(self, instance: object, *values) -> None:
    """Dispatches the pre-del notifications."""
    for callMeMaybe in self._preDelCallbacks():
      callMeMaybe(instance, *values)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Check presence of notification callbacks

  def hasOnSetCallbacks(self, ) -> bool:
    """Returns True if there are any on-set callbacks."""
    for _ in self._onSetCallbacks():
      return True
    return False

  def hasPreSetCallbacks(self, ) -> bool:
    """Returns True if there are any pre-set callbacks."""
    for _ in self._preSetCallbacks():
      return True
    return False

  def hasOnDelCallbacks(self, ) -> bool:
    """Returns True if there are any on-del callbacks."""
    for _ in self._onDelCallbacks():
      return True
    return False

  def hasPreDelCallbacks(self, ) -> bool:
    """Returns True if there are any pre-del callbacks."""
    for _ in self._preDelCallbacks():
      return True
    return False

  def hasPreGetCallbacks(self, ) -> bool:
    """Returns True if there are any pre-get callbacks."""
    for _ in self._preGetCallbacks():
      return True
    return False

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Implementation of the descriptor protocol

  def __get__(self, instance: object, owner: type, **kwargs) -> Any:
    """Get the value of the field."""
    if instance is None:
      return self
    try:
      value = self.__instance_get__(instance)
    except Exception as exception:
      if kwargs.get('_setterAsks', False):
        return None
      else:
        raise exception
    if kwargs.get('_setterAsks', False):
      return value
    if self.hasPreGetCallbacks():
      self.notifyPreGet(instance, value)
    return value

  def __set__(self, instance: object, value: object) -> None:
    """Set the value of the field."""
    if not (self.hasPreSetCallbacks() or self.hasOnSetCallbacks()):
      return self.__instance_set__(instance, value)
    oldValue = self.__get__(instance, type(instance), _setterAsks=True)
    if self.hasPreSetCallbacks():
      self.notifyPreSet(instance, oldValue, value)
    self.__instance_set__(instance, value)
    self.notifyOnSet(instance, oldValue, value)

  def __delete__(self, instance: object) -> None:
    """Delete the value of the field."""
    if not (self.hasPreDelCallbacks() or self.hasOnDelCallbacks()):
      return self.__instance_del__(instance)
    oldValue = self.__get__(instance, type(instance), _setterAsks=True)
    self.notifyPreDel(instance, oldValue)
    self.__instance_del__(instance)
    self.notifyOnDel(instance, oldValue)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Instance-specific accessor methods

  @abstractmethod
  def __instance_get__(self, instance: object) -> Any:
    """Subclasses should implement this method to provide the getter. """

  def __instance_set__(self, instance: object, value: object) -> None:
    """Subclasses should implement this method to provide the setter. """
    descName = self.__class__.__name__
    ownerName = self.getFieldOwner().__name__
    fieldName = self.getFieldName()
    e = """The attribute '%s' on class '%s' belongs to descriptor of type: 
    '%s' which does not implement setting!"""
    raise TypeError(monoSpace(e % (fieldName, ownerName, descName)))

  def __instance_del__(self, instance: object) -> None:
    """Subclasses should implement this method to provide the deleter. """
    descName = self.__class__.__name__
    ownerName = self.getFieldOwner().__name__
    fieldName = self.getFieldName()
    e = """The attribute '%s' on class '%s' belongs to descriptor of type: 
    '%s' which does not implement deletion!"""
    raise TypeError(monoSpace(e % (fieldName, ownerName, descName)))



"""FastMeta provides the metaclass for the FastData class. This metaclass
customize the initial class creation process by returning an instance of
FastSpace from __prepare__ for use as namespace. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import BaseMetaclass, Bases
from worktoy.base import FastSpace


class FastMeta(BaseMetaclass):
  """FastMeta provides the metaclass for the FastData class. This metaclass
  customize the initial class creation process by returning an instance of
  FastSpace from __prepare__ for use as namespace. """

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> FastSpace:
    return FastSpace(mcls, name, bases, **kwargs)



"""BaseMetaclass provides general functionality for derived classes. This
includes primarily function overloading. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object

from worktoy.meta import Bases, OverloadSpace, AbstractMetaclass
from worktoy.text import monoSpace, typeMsg


class BaseMetaclass(AbstractMetaclass):
  """BaseMetaclass provides general functionality for derived classes. This
  includes primarily function overloading. """

  def __instancecheck__(cls, instance: object) -> bool:
    """The __instancecheck__ method is called when the 'isinstance' function
    is called."""
    if getattr(cls, '__class_instancecheck__', None) is None or True:
      return AbstractMetaclass.__instancecheck__(cls, instance)
    instanceCheck = getattr(cls, '__class_instancecheck__', )
    if not callable(instanceCheck):
      e = typeMsg('instanceCheck', instanceCheck, Callable)
      raise TypeError(monoSpace(e))
    if not isinstance(instanceCheck, classmethod):
      e = """The instanceCheck method must be a classmethod!"""
      e2 = typeMsg('instanceCheck', instanceCheck, classmethod)
      raise TypeError(monoSpace("""%s %s""" % (e, e2)))
    if getattr(instanceCheck, '__self__', None) is None:
      return instanceCheck(cls, instance)
    return instanceCheck(instance)

  def __subclasscheck__(cls, subclass) -> bool:
    """The __subclasscheck__ method is called when the 'issubclass' function
    is called."""
    if getattr(cls, '__class_subclasscheck__', None) is None or True:
      return AbstractMetaclass.__subclasscheck__(cls, subclass)
    subclassCheck = getattr(cls, '__class_subclasscheck__', )
    if not callable(subclassCheck):
      e = typeMsg('subclassCheck', subclassCheck, Callable)
      raise TypeError(monoSpace(e))
    if not isinstance(subclassCheck, classmethod):
      e = """The subclassCheck method must be a classmethod!"""
      e2 = typeMsg('subclassCheck', subclassCheck, classmethod)
      raise TypeError(monoSpace("""%s %s""" % (e, e2)))
    if getattr(subclassCheck, '__self__', None) is None:
      return subclassCheck(cls, subclass)
    return subclassCheck(subclass)

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> OverloadSpace:
    """The __prepare__ method is invoked before the class is created. This
    implementation ensures that the created class has access to the safe
    __init__ and __init_subclass__ through the BaseObject class in its
    method resolution order."""
    return OverloadSpace(mcls, name, bases, **kwargs)

  def __new__(mcls,
              name: str,
              bases: Bases,
              space: OverloadSpace,
              **kwargs) -> type:
    """The __new__ method is invoked to create the class."""
    namespace = space.compile()
    if '__del__' in namespace and '__delete__' not in namespace:
      if not kwargs.get('trustMeBro', False):
        e = """The namespace encountered the '__del__' method! 
          This method has very limited practical use. It has significant 
          potential for unexpected behaviour. Because the '__del__' method 
          were implemented, but not the '__delete__' method, this error
          was raised. If '__del__' were the intention, please provide the 
          keyword 'trustMeBro=True' to the class creation."""
        raise AttributeError(monoSpace(e))
    return AbstractMetaclass.__new__(mcls, name, bases, namespace, **kwargs)
